//Zero Dimensional PONG
//2024 by SENIL DATA SYSTEMS

//consider this code free-for-all. (although replacing my name with yours and selling it as your code would be a dick move.

//my intention is to make a zero-dimensional PONG game.
//You know PONG. If not, go ahead and put TinyPONG on an Atmega8, hook up two potentiometers and a HD44780 character LCD and you'll quickly get the point.
//So, people have built one-dimensional PONG games - removing one dimension and have a single lit LED bounce around a string of LEDs and each player has to
//press the button at the right time to hit the ball.
//Now I'm thinking that's still one dimension too many.
//You can substitute the position of the ball on the line through Pulse Width Modulation.
//And since it's quite easy to spot when the LED is off, it's hard to spot when it's 100% on, thus I'm gonna wire this up for two-color LEDs.
//If you don't like that, just hook the other pin of the LED to ground instead of whatever output pin I decide upon.

//requirements...
//Two LEDs -> two pins
//Two buttons -> two pins
//A piezo speaker - one pin
//a 3 digit LED display - 7+4 pins -> eleven pins (I could use multiplex ICs, but why?)
//that's 16 pins. Ugh. The Attiny2313 has 17 or 18.
//so how about display column 0,1,2,3 go to PD0,1,4,5
//and display segments go to PORTB minus PB3
//a-0 b-1 c-2 d-4 e-5 f-6 g-7 (because 3 is beeper)

//so I think I'm gonna hook the two buttons up to the interrupt pins.
//My plan is to do the PWM by busy waiting even though the Atmega family has hardware PWM.

//coming from TinyPONG, I'll have to use OC1A for the beeper. That's pin 15. That sucks. I'd like to have all 7 bits to the 7 segment display as one byte...
//hey wait a minute - the entire 8 bits of PortB is available. I'd just need to write around that. And the beeper might make weird noises when the display strobes.
//the buttons connect to the interrupts, so that's pin 6 and 7.
//let's hook the LEDs up to PA0 and PA1 (pins 4 and 5)

/*
Lemme tell you a little story...
I was testing interrupts. Simple deal, just set direction=-1 on IRQ0 and direction=1 on IRQ1. Simple.
But it didn't work, it looked like the interrupts never executed.
I added speed++ and speed-- in the ISRs which didn't affect anything either.
Then I made IRQ0 set all PORTB to on and IRQ1 to set them off and was all confused as the ISRs actually did run after all.
Then I added sound, which worked.
Then I made one ISR to output speed to PORTB and the other to output bright to PORTB.
While you could see speed nicely counting up scoping the BUS, bright always output 1.
So I set speed to 0x55 - and that's what I got.
So... the ISR and the main loop had two copies of the same variables and each used their own!
So I tried static. That didn't do diddly.

Then I removed compiler optimization. And now everything works. WTF?
Well unfortunately, without optimization, even the hello world balloons out and barely fits the Flash.
Reading in the listing, you can see that in the main loop, it never loads the value of direction from RAM
instead it uses a constant. Since direction is only set once in main, the optimizations screw that up.

The solution was to use the volatile keyword.
You know what's weird? I didn't have that problem with tapedeck.c

But now I have the same problem that triggering IRQ0 often makes the ISR add two instead of one.
*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <util/delay.h>

static volatile unsigned char bright;
static volatile char direction;
static volatile unsigned char speed;
static volatile unsigned char p1score;
static volatile unsigned char p2score;
static volatile unsigned char gamestate; //let's define 0=game over, 1=difficulty setting, 2 game running (maybe 2..255 when to speed up)
static volatile unsigned char difficulty;
static volatile unsigned char paddleflag; //there's still the bug if you hit BOTH buttons like crazy, it can cause the same bug where both players get points super quickly, but at least it's fixed if any one player is spamming the button.
static volatile unsigned char sound; //so we don't have to put delay_ms in our ISRs.
/*
let's define some difficulties
0: Boring. Speed very slow, very slowly ramps up, hit range is pretty liberal, like say 32.
1: Hit range 32, speed depends on how close to home you hit it, max speed is medium
2,3 : same as 0,1 with hit range 20, speed slow, max speed fast
4,5 : same as 0,1 with hit range 10, speed medium, max speed very fast
6,7 : same as 0,1 with hit range 4, speed medium, max speed = as fast as I can make it go without changing the PWM loop to be shorter
8,9 : same as 0,1 with hit range 1, speed slightly fast.

You know I *could* make this zero dimensional Tennis for Two - 
if you push the button before the ball is on your side of the screen, it is ignored.
Otherwise, the speed of the ball depends on how far you let it get to your side.
So if you push the button way early, the ball will be so slow, the opponent will easily get it.
But if you wait until the ball is almost out of the court, KABLAM! Your opponent won't know what hit him.
(no, I'm not gonna do that, I'll stick with regular PONG. Minus two dimensions.)
*/

uint8_t sevensegment[]={ //note that PB3 can't be used
	//gfed1cba
	0b01111111, //0
	0b00001110, //1
	0b10111011, //2
	0b10011111, //3
	0b11001110, //4
	0b11011101, //5
	0b11111101, //6
	0b00001111, //7
	0b11111111, //8
	0b11011111, //9
	0b10111110  //d (for "difficulty")
};

uint8_t hitrange[]={ //tweak this to your liking
	40,
	28,
	15,
	8,
	4 //OK I tried. On level 6 I could hit it once every couple of times.
};

uint8_t speedlist[]={ //tweak this to your liking
	20,
	16, //this takes the ball very roughly the same amount of time from side to side as regular PONG on the AY-3-8500.
	12,
	12,
	8
};


//stolen from PONG, probably copied off the internet somewhere. That's why the TCCR1B is in there too.
void timer1_init(void)
{
    // set up timer with prescaler = 8 and CTC mode
    TCCR1B |= (1 << WGM12)|(1 << CS11);//|(1 << CS10);
 
    // set up timer OC1A pin in toggle mode
    TCCR1A |= (1 << COM1A0);
 
    // initialize counter
    TCNT1 = 0;
    // initialize compare value
    //OCR1A = 24999;
}


int main(void)
{
	//well first let's just make an LED blink.
	DDRA = 	255; //all 8 pins are output...
	DDRB =  255;
	//there's no PortC on this thing.
	DDRD =  0b11110011; //configure IRQ0 and 1 as inputs
	PORTD = 0b00001100; //writing 1 to a pin set as input enables internal pull-up resistors
	unsigned char i=0;
	unsigned char c;
	bright=0x55; //wanna bet?
	direction=1; //can be 1 or -1
	speed=16; //the lower, the faster
	c=0;
	p1score=0;
	p2score=0;
	gamestate=1;
	difficulty=0;
	
	
    //#MCUCR |= (1 << ISC10);
    MCUCR |= (1 << ISC11); //ok if I outcomment that, the machine halts when I hold a button pressed, but won't execute the interrupt.

    //#MCUCR |= (1 << ISC00);
    MCUCR |= (1 << ISC01);
	
	//isc00 0 && isc01 0 -> int0 on logical low
	//isc00 1 && isc01 0 -> int0 on any change
	//isc00 0 && isc01 1 -> int0 on falling edge
	//isc00 1 && isc01 1 -> int0 on rising edge
	//the most fun thing about programming through the Atmega/tiny family is that for each variant of the chip, this register is named different.
	GIMSK |= (1 << INT1);                // Enable INT1
    GIMSK |= (1 << INT0);                // Enable INT0
    EIFR |= (1 << INTF1);                // Enable INT1
    EIFR |= (1 << INTF0);                // Enable INT0
	//on Atmega48, the register is named differently
	//EIMSK |= (1 << INT1);                // Enable INT1
    //EIMSK |= (1 << INT0);                // Enable INT0


	sei();
	timer1_init();
	
	OCR1A= 600;
	TCCR1A |= (1 << COM1A0); //sound on
	_delay_ms(50); //how long you want to beep - note this shuts the display and PWM down unless I make them an interrupt. Since sounds can only happen at or near 0% / 100%, the main LED is not a problem.s
	TCCR1A=0;

	//strobe the display
	TCCR0B = 2; //4=(/1024) 3=(/64) 2=(/8) 1=run at 8MHz - so use two if it's flickering
	TIMSK |= (1<<TOIE0);

	
	//note _delay_ms(10) comes out as more around 17-18ms if I disable CLKDIV8.
	while(1){
		for(i=0; i<255; i++){ //ok it does one loop through the entire range. If that's too flickery, I need it to count only to say 127 or less.
			if(i>bright) 	PORTA = 1;
			else 			PORTA = 2;
		}
		if(c==speed) {
			//this needs to be here - the position of the ball is added immediately hereafter
			//so if it's at 255 and moving "up", or at 0 and moving "down", it's gonna wrap NOW, 
			//meaning the player has missed the ball
			if(gamestate>1){
				if(bright>254 && direction<128) { //another gotcha - you can't check if direction < 0. You must check if it's 255.
					p2score++;
					if(p2score==10) p2score=0x10;
					bright=1; //reset the ball - otherwise hitting the ball *while* the goal sound plays is a legal move!
					OCR1A= 120; //high note denotes goal
					TCCR1A |= (1 << COM1A0); //sound on
					_delay_ms(20);
					TCCR1A=0;
				}
				if(bright<1 && direction>128) {
					p1score++;
					if(p1score==10) p1score=0x10;
					bright=254;
					OCR1A= 120; //high note denotes goal
					TCCR1A |= (1 << COM1A0); //sound on
					_delay_ms(20);
					TCCR1A=0;
					speed=speedlist[difficulty>>1]; //reset speed after goal
				}
				if(p1score>=0x15 || p2score>=0x15) {
					gamestate=0; //game over
					OCR1A= 1200; //middle note denotes ball hit
					TCCR1A |= (1 << COM1A0); //sound on
					_delay_ms(200);
					TCCR1A=0;
					speed=speedlist[difficulty>>1]; //reset speed after goal
				}
			}
			
			bright+=direction; //IT DOESN'T LOAD THE VALUE OF direction FROM MEMORY! Instead it just uses a static value (and 0xFF for that)
			if(bright==128) paddleflag=0; //|direction| must never exceed 1 (at 4 it's fast as f#*k, so it won't - unless you check that CKDIV8)
			c=0;
		}
		c++;
		if(sound>0) sound--;
		else TCCR1A=0;
		
		//ugh now we gotta define which player is which button. 
		//INT0 button is right and blue. The right pin of the LED should be blue or green and PA0
		//INT1 button is left and red. The left pin of the LED should be red and PA1.
		//so direction=1 is green and goes to red, so if bright=>255-[difficulty margin], it's a score for player 2
		//so direction=-1 is red and goes to green, so if bright<=[difficulty margin], it's a score for player 1

		//HELLO WORLD - just blink an LED
		//PORTA = 1;
		//_delay_ms(100);
		//PORTA = 2;
		//_delay_ms(100);
		
		//if I read the buttons here, they work "even" with optimization on
		
	}


	return 0;
	
}

int digitalRead(int pin) //works but exhibits strange behaviour sometimes (on startup, always reports H for a while, until you really FORCE the signal down for quite a while, after that, it worked just fine, but this happens rarely and is probably my fault.)
{
	//pin--;
	//ps2_gohi(pin); //sollte eigentlich hochohmig sein...
	return PIND & (1<<pin);
}


//OK problem - unless I don't optimize at all, the ISR doesn't use the global variables and is thus useless.
//but if I don't optimize at all, even the simplest code will blow the 2k capacity!
ISR(INT0_vect){
	//if bright<difficulty //note this will allow a player to quickly push the button and get lucky
	/*DEBUG CODE
	direction=1;
	//else if bright<128 set flag to disable the interrupt which prevents the player from hitting the ball
	
	speed++;
	PORTB=bright;

	OCR1A= 200;
	TCCR1A |= (1 << COM1A0); //sound on
	_delay_ms(20); //yeah yeah yeah delays in interrupts... don't try that at home, kiddo!
	TCCR1A=0;*/
	unsigned char temp;
	if((paddleflag&1)==1) return;
	cli();
	if(gamestate==1){ //ok I have the same problem as in tapedeck - this adds 2 pretty much always.
		difficulty++;
		if(difficulty>18) difficulty=0; //OK so you're adding 2 almost always, well that's your choice and I gotta work with it.
		_delay_ms(5); //poor man's debounce
		temp=difficulty>>1;
		speed=speedlist[temp>>1];
		bright=32;
		direction=1;

	} else if(gamestate>=2){ //FIXME: For both players - you can repeatedly press the button while the ball is in the players' catch zone and score one point for both players at the same time... the other ISR has the same bug.
								//my plan would be to have a flag for each player that he hit the ball. The flag is reset when the ball crosses the midpoint.
		//check if ball hit
		paddleflag=1;
		if(bright<=hitrange[difficulty>>1] && direction>128){
			direction=1;
			OCR1A= 250; //middle note denotes ball hit
			TCCR1A |= (1 << COM1A0); //sound on
			sound=100;
			
			if((difficulty&1)==0){
				gamestate++; //ball gets faster after successive hits
				if(gamestate>8){
					gamestate=2;
					if(speed>2) speed--;
				}
			} else {
				speed=bright;
				temp=speedlist[difficulty>>1]>>1;
				if(speed<temp) speed=temp; //make sure it doesn't go faster than twice the base setting
				if(speed>32) speed=32; //and make sure it doesn't go too slow (just a sanity check)
			}

		} else if (bright<64 && direction>128) {
			p1score++;
			if(p1score==10) p1score=0x10;
			bright=222;
			OCR1A= 150; //off high note denotes player 2 hit early
			TCCR1A |= (1 << COM1A0); //sound on
			sound=100;

		}
	}
	sei();
}

//LATCH PD3 Pin 5
ISR(INT1_vect){
	//if bright>255-difficulty //see note in IRQ0
	/* DEBUG CODE
	direction=-1;
	
	speed--; //DEBUG

	OCR1A= 400;
	TCCR1A |= (1 << COM1A0); //sound on
	_delay_ms(20);
	TCCR1A=0;
	*/
	unsigned char temp; //hello compiler, this is not volatile and you can optimize it away from memory and put it in a CPU register instead.
	if((paddleflag&2)==2) return;
	if(gamestate==1) { //START GAME
		gamestate=2;
		difficulty>>=1; //ok if I can't prevent it from adding 2 most of the time, I'll expect it to add 2 always.
		OCR1A= 1200; //"deep" notes announces start of game
		TCCR1A |= (1 << COM1A0); //sound on
		sound=100;

	} else if(gamestate>=2){
		paddleflag=2;
		//check if ball hit
		if(bright>(255-hitrange[difficulty>>1]) && direction<127){
			OCR1A= 250; //middle note denotes ball hit
			TCCR1A |= (1 << COM1A0); //sound on
			sound=100;
			direction=-1;

			//add speed if ball is kept in play
			if((difficulty&1)==0){
				gamestate++;
				if(gamestate>8){ //on even difficulties, every 6 hits
					gamestate=2;
					if(speed>2) speed--;
				}
			} else { //on odd difficulties, speed depends on when you hit it.
				speed=255-bright;
				temp=speedlist[difficulty>>1]>>1;
				if(speed<temp) speed=temp; //make sure it doesn't go faster than twice the base setting
				if(speed>32) speed=32; //and make sure it doesn't go too slow
			}
		} else if (bright>192 && direction<127){
			bright=32;
			p2score++;
			if(p2score==10) p2score=0x10;
			OCR1A= 150; //off high note denotes player 1 hit early
			TCCR1A |= (1 << COM1A0); //sound on
			sound=100;
		}
	
	}// doesn't work right. Should go to difficulty selection but starts new game instead.
	else if(gamestate==0){ //if GAME OVER and both buttons pressed, start new game
		paddleflag=2;
		if(digitalRead(2)==0){
			p1score=0;
			p2score=0;
			gamestate=1;
		}
	}
}

//and the display. I'll have it count to 15 even though this is not a 4-bit console as are the original PONG machines.
ISR(TIMER0_OVF_vect){

	static unsigned char colstrobe=0;
	PORTD &= 0b00001100;
	PORTB &= 0b00001000;
	switch(colstrobe){
		case 0:
			//segment 1, a=. and d=:
			PORTD |= 1;
			colstrobe++;
			if(gamestate==1){ //difficulty
				PORTB |= ~1; //light up decimal point
			} else {
				PORTB |= ~16; //light up colon
				if(p1score>=16) PORTB &= ~6; //leading one
			}
			break;
		case 1:
			PORTD |= 2;
			colstrobe++;
			if(gamestate==1){
				PORTB|= ~sevensegment[10];
			}else{
				PORTB|= ~sevensegment[p1score&15];
			}
			break;
		case 2:
			PORTD |= 16;
			colstrobe++;
			if(gamestate==1){
				PORTB|= ~sevensegment[difficulty>>1]; //this way this can display higher numbers than 1
			} else {
				if(p2score>=16) PORTB |= ~6; else PORTB=255;//leading one
			}
			break;
		case 3:
			PORTD |= 32;
			colstrobe=0;
			if(gamestate!=1){
				PORTB|= ~sevensegment[p2score&15];
			} else {
				PORTB=255;
			}
			break;
		default:
			colstrobe=0;
	}

}